@@ -28,7 +28,8 @@ OTHER DEALINGS IN THE SOFTWARE.
 For more information, please refer to <http://unlicense.org/>
 """
 
-from typing import Iterable, List, NamedTuple, Optional, Union
+from json import loads
+from typing import Iterable, List, NamedTuple, Optional, Type, Union
 
 from tempfile import TemporaryDirectory
 from argparse import ArgumentParser
@@ -38,16 +39,34 @@ from pathlib import Path
 import subprocess
 
 
+QUERY_CLIP_START = "  -   clip end: "
+QUERY_CLIP_END = "  - clip start: "
+QUERY_CLIP_STATUS = "  -     status: "
+
+MSG_INFO = ""
+MSG_DEBUG = ""
+MSG_ERROR = ""
+
+_FF_QUIET = "-hide_banner -loglevel error"
+FF_QUIET: List[str] = _FF_QUIET.split()
+CLIP_FFARGS: str = _FF_QUIET + (
+    " -loop 1 -c:a aac -vcodec libx264 -pix_fmt yuv420p -preset ultrafast -tune "
+    "stillimage -shortest"
+)
+
+
 class Behaviour(NamedTuple):
     """
     typed command line argument tuple
     """
 
     query: str
-    dir: Optional[Path]
+    dir: Path
     out: Optional[Path]
-    sdargs: str
+    sdargs: List[str]
+    ffargs: List[str]
     clip_length: int
+    use_defaults: bool
 
 
 def main() -> None:
@@ -62,82 +81,123 @@ def main() -> None:
         tmpdir = Path(_tmpdir)
 
         # download songs
-        invocate(
-            name="spotdl", args=[bev.query] + bev.sdargs.split(), cwd=tmpdir, errcode=2
-        )
+        invocate(name="spotdl", args=[bev.query] + bev.sdargs, cwd=tmpdir, errcode=2)
 
         # process songs
-        print("\npymtheg: info: enter timestamps in format [hh:mm:]ss")
-        print("               end timestamp can be relative, prefix with '+'")
-        print("               press enter to use given defaults")
+        if bev.use_defaults:
+            print(
+                "\npymtheg: info: using defaults, clip start will be 0 and clip end will"
+                f" be {bev.clip_length}"
+            )
 
-        for song_path in tmpdir.rglob("*.mp3"):
-            print(f"  {song_path.stem}")
+        else:
+            print("\npymtheg: info: enter timestamps in format [hh:mm:]ss")
+            print("               end timestamp can be relative, prefix with '+'")
+            print("               press enter to use given defaults")
 
-            # get starting timestamp
+        for song_path in tmpdir.rglob("*.*"):
+            # ensure that file was export of spotDL (list from spotdl -h)
+            if song_path.suffix not in [".m4a", ".ogg", ".flac", ".mp3", ".wav", ".opus"]:
+                continue
+
+            print(f"- {song_path.stem}")
+
+            _msg = f"{QUERY_CLIP_STATUS}probe song duration"
+            print(_msg, end="\r")
+
+            # duration retrieval
+            proc = invocate(
+                "ffprobe",
+                args=[
+                    "-print_format",
+                    "json",
+                    "-show_entries",
+                    "format=duration",
+                    song_path,
+                ],
+                capture_output=True,
+            )
+            song_duration: int = int(
+                loads(proc.stdout)["format"]["duration"].split(".")[0]
+            )
+
+            print(" " * len(_msg), end="\r")
+
+            # get timestamps
             start_timestamp = 0
-
-            _query = "    clip start: "
-            while True:
-                response = input(f"{_query}0\r{_query}")
-
-                if response != "":
-                    _start_timestamp = parse_timestamp(response)
-
-                    if _start_timestamp is None:
-                        # reprompt if invalid
-                        print(
-                            (" " * len(_query)) + ("^" * len(response)),
-                            "invalid timestamp",
-                        )
-
-                    else:
-                        start_timestamp = _start_timestamp
-                        break
-
-                elif response == "":
-                    break
-
-            # get ending timestamp
             end_timestamp: int = start_timestamp + bev.clip_length
 
-            _query = "      clip end: "
-            while True:
-                response = input(f"{_query}+{bev.clip_length}\r{_query}")
-                if response != "":
-                    _end_timestamp = parse_timestamp(
-                        response, relative_to=start_timestamp
-                    )
+            if not bev.use_defaults:
+                # starting timestamp
+                while True:
+                    response = input(f"{QUERY_CLIP_START}0\r{QUERY_CLIP_START}")
 
-                    if _end_timestamp is None:
-                        # reprompt if invalid
-                        print(
-                            (" " * len(_query)) + ("^" * len(response)),
-                            "invalid timestamp",
-                        )
+                    if response != "":
+                        _start_timestamp = parse_timestamp(response)
 
-                    else:
-                        end_timestamp = _end_timestamp
+                        if _start_timestamp is None:
+                            # invalid format
+                            print(
+                                (" " * len(QUERY_CLIP_START)) + ("^" * len(response)),
+                                "invalid timestamp",
+                            )
+
+                        elif _start_timestamp >= song_duration:
+                            # invalid, timestamp >= song duration
+                            print(
+                                (" " * len(QUERY_CLIP_START)) + ("^" * len(response)),
+                                "timestamp exceeds song duration",
+                            )
+
+                        else:
+                            # valid, continue
+                            start_timestamp = _start_timestamp
+                            break
+
+                    elif response == "":
                         break
 
-                elif response == "":
-                    break
+                # ending timestamp
+                while True:
+                    response = input(
+                        f"{QUERY_CLIP_END}+{bev.clip_length}\r{QUERY_CLIP_END}"
+                    )
+                    if response != "":
+                        _end_timestamp = parse_timestamp(
+                            response, relative_to=start_timestamp
+                        )
+
+                        if _end_timestamp is None:
+                            # reprompt if invalid
+                            print(
+                                (" " * len(QUERY_CLIP_END)) + ("^" * len(response)),
+                                "invalid timestamp",
+                            )
+
+                        else:
+                            end_timestamp = _end_timestamp
+                            break
+
+                    elif response == "":
+                        break
 
             # construct paths
             song_path = song_path.absolute()
             song_clip_path = tmpdir.joinpath(f"{song_path.stem}_clip.mp3").absolute()
             song_cover_path = tmpdir.joinpath(f"{song_path.stem}_cover.png").absolute()
-            out_path: Path = Path(f"{song_path.stem}.mp4")
+            out_path: Path = bev.dir.joinpath(f"{song_path.stem}.mp4").absolute()
 
             if bev.out is not None:
                 out_path = bev.out
-            elif bev.dir is not None:
-                out_path = bev.dir.joinpath(f"{song_path.stem}.mp4")
 
             # clip audio
+            _msg = f"{QUERY_CLIP_STATUS}clip audio"
+            print(_msg, end="\r")
+
             invocate(
                 "ffmpeg",
                 args=[
+                    *FF_QUIET,
                     "-ss",
                     str(start_timestamp),
                     "-to",
@@ -150,10 +210,16 @@ def main() -> None:
                 errcode=3,
             )
 
+            print(" " * len(_msg), end="\r")
+
             # get album art
+            _msg = f"{QUERY_CLIP_STATUS}getting album art"
+            print(_msg, end="\r")
+
             invocate(
                 "ffmpeg",
                 args=[
+                    *FF_QUIET,
                     "-i",
                     song_path,
                     "-an",
@@ -163,29 +229,14 @@ def main() -> None:
                 errcode=3,
             )
 
-            # make video
+            print(" " * len(_msg), end="\r")
+
+            # creating clip
+            print(f"{QUERY_CLIP_STATUS}creating clip")  # no \r because ffmpeg output
+
             invocate(
                 "ffmpeg",
-                args=[
-                    "-loop",
-                    "1",
-                    "-i",
-                    song_cover_path,
-                    "-i",
-                    song_clip_path,
-                    "-c:a",
-                    "aac",
-                    "-vcodec",
-                    "libx264",
-                    "-pix_fmt",
-                    "yuv420p",
-                    "-preset",
-                    "ultrafast",
-                    "-tune",
-                    "stillimage",
-                    "-shortest",
-                    out_path,
-                ],
+                args=["-i", song_cover_path, "-i", song_clip_path, *bev.ffargs, out_path],
                 errcode=3,
             )
 
@@ -248,6 +299,7 @@ def invocate(
     args: Iterable[Optional[Union[str, Path]]] = [],
     cwd: Optional[Path] = None,
     errcode: int = -1,
+    capture_output: bool = False,
 ) -> subprocess.CompletedProcess:
     """
     invocates command using subprocess.run
@@ -256,7 +308,7 @@ def invocate(
         name of program
     args: Iterable[Optional[Union[str, Path]]] = [],
         args of program, e.g. ["download", "-o=$HOME"]
-    wd: Optional[Path] = None,
+    cwd: Optional[Path] = None,
         working directory for process to be run
     errcode: int = -1,
         exit code for if the process returns non-zero
@@ -269,10 +321,26 @@ def invocate(
             invocation.append(arg)
 
     try:
-        print(
-            f"pymtheg: info: invocating command '{' '.join([str(p) for p in invocation])}'\n"
+        proc = subprocess.run(
+            invocation,
+            cwd=cwd,
+            universal_newlines=True,
+            capture_output=capture_output,
         )
-        return subprocess.run(invocation, cwd=cwd, universal_newlines=True, check=True)
+
+        if proc.returncode != 0:
+            if capture_output:
+                if proc.stdout != "":
+                    print(f"\npymtheg: error: invocation stdout:\n{proc.stdout}")
+                if proc.stderr != "":
+                    print(f"\npymtheg: error: invocation stderr:\n{proc.stderr}")
+
+            print(
+                "\npymtheg: error: error during invocation of "
+                f"'{' '.join([str(p) for p in invocation])}', returned non-zero exit "
+                f"code {proc.returncode}, see above for details"
+            )
+            exit(proc.returncode)
 
     except FileNotFoundError as err:
         print_tb(err.__traceback__)
@@ -282,14 +350,6 @@ def invocate(
         )
         exit(errcode)
 
-    except subprocess.CalledProcessError as err:
-        print_tb(err.__traceback__)
-        print(
-            f"{err.__class__.__name__}: {err}\n\n"
-            f"pymtheg: error: error during invocation of {name}, see traceback"
-        )
-        exit(errcode)
-
     except Exception as err:
         print_tb(err.__traceback__)
         print(
@@ -298,6 +358,9 @@ def invocate(
         )
         exit(errcode)
 
+    else:
+        return proc
+
 
 def get_args() -> Behaviour:
     """
@@ -306,12 +369,14 @@ def get_args() -> Behaviour:
     # parse
     parser = ArgumentParser(
         prog="pymtheg",
-        description="a python script to share songs from Spotify/YouTube as a 15 second clip",
+        description=(
+            "a python script to share songs from Spotify/YouTube as a 15 second clip"
+        ),
     )
 
     parser.add_argument("query", help="song/link from spotify/youtube")
     parser.add_argument(
-        "-d", "--dir", type=Path, help="directory to output to", default=None
+        "-d", "--dir", type=Path, help="directory to output to", default=Path.cwd()
     )
     parser.add_argument(
         "-o",
@@ -321,6 +386,12 @@ def get_args() -> Behaviour:
         default=None,
     )
     parser.add_argument("-sda", "--sdargs", help="args to pass to spotdl", default="")
+    parser.add_argument(
+        "-ffa",
+        "--ffargs",
+        help="args to pass to ffmpeg for clip creation",
+        default=CLIP_FFARGS,
+    )
     parser.add_argument(
         "-cl",
         "--clip-length",
@@ -329,30 +400,44 @@ def get_args() -> Behaviour:
         type=int,
         default=15,
     )
+    parser.add_argument(
+        "-ud",
+        "--use-defaults",
+        help="use 0 as clip start and --clip-length as clip end",
+        dest="use_defaults",
+        action="store_true",
+        default=False,
+    )
 
     args = parser.parse_args()
     bev = Behaviour(
         query=args.query,
         dir=args.dir,
         out=args.out,
-        sdargs=args.sdargs,
+        sdargs=args.sdargs.split(),
+        ffargs=args.ffargs.split(),
         clip_length=args.clip_length,
+        use_defaults=args.use_defaults,
     )
 
-    # validate
-    if bev.out is not None and bev.out.exists():
-        override_response = input(f"pymtheg: info: {bev.out} exists, override? (y/n)")
-        if override_response.lower() != "y":
+    # validate:
+    if bev.out is not None:
+        if bev.out.is_dir():
+            print("pymtheg: error: output file is a directory")
             exit(1)
 
-    if bev.dir is not None:
-        if not bev.dir.exists():
-            print("pymtheg: error: output directory is non-existent")
-            exit(1)
+        if bev.out.exists():
+            override_response = input(f"pymtheg: info: {bev.out} exists, override? (y/n)")
+            if override_response.lower() != "y":
+                exit(1)
 
-        if not bev.dir.is_dir():
-            print("pymtheg: error: output directory is not a directory")
-            exit(1)
+    if not bev.dir.exists():
+        print("pymtheg: error: output directory is non-existent")
+        exit(1)
+
+    if not bev.dir.is_dir():
+        print("pymtheg: error: output directory is not a directory")
+        exit(1)
 
     return bev
 
